@@ -1,7 +1,7 @@
 #  File src/library/base/R/library.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2014 The R Core Team
+#  Copyright (C) 1995-2020 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -14,27 +14,74 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
-testPlatformEquivalence <-
-function(built, run)
-{
-    ## args are "cpu-vendor-os", but os might be 'linux-gnu'!
-    ## remove vendor field
-    built <- gsub("([^-]*)-([^-]*)-(.*)", "\\1-\\3", built)
-    run <- gsub("([^-]*)-([^-]*)-(.*)", "\\1-\\3", run)
-    ## Mac OS X supports multiple CPUs by using 'universal' binaries
-    if (grepl("^universal-darwin", built) && nzchar(.Platform$r_arch))
-        built <- sub("^universal", R.version$arch, built)
-    ## allow for small mismatches, e.g. OS version number and i686 vs i586.
-    length(agrep(built, run)) > 0
-}
+## Usage removed in 3.6.0
+## testPlatformEquivalence <-
+## function(built, run)
+## {
+##     ## args are "cpu-vendor-os", but os might be 'linux-gnu'!
+##     ## remove vendor field
+##     built <- gsub("([^-]*)-([^-]*)-(.*)", "\\1-\\3", built)
+##     run <- gsub("([^-]*)-([^-]*)-(.*)", "\\1-\\3", run)
+##     ## macOS supports multiple CPUs by using 'universal' binaries
+##     if (startsWith(built, "universal-darwin") && nzchar(.Platform$r_arch))
+##         built <- sub("^universal", R.version$arch, built)
+##     ## allow for small mismatches, e.g. OS version number and i686 vs i586.
+##     length(agrep(built, run)) > 0
+## }
+
+## If we want this it would be better to factor out the core of checkConflicts.
+## searchConflicts <- function(pkg) {
+##     vars <- getNamespaceExports(pkg)
+##     conflicts <- function(pos) intersect(vars, ls(pos, all.names = TRUE))
+##     val <- Filter(length, sapply(search()[-1], conflicts))
+##     if (length(val)) val else NULL
+## }
+
+conflictRules <-
+    local({
+        data <- new.env()
+        function(pkg, mask.ok = NULL, exclude = NULL) {
+            if ((! missing(mask.ok)) || (! missing(exclude)))
+                assign(pkg, list(mask.ok = mask.ok, exclude = exclude),
+                       envir = data)
+            else
+                get0(pkg, envir = data, inherits = FALSE)
+        }
+    })
 
 library <-
 function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
-         logical.return = FALSE, warn.conflicts = TRUE,
-	 quietly = FALSE, verbose = getOption("verbose"))
+         logical.return = FALSE, warn.conflicts,
+	 quietly = FALSE, verbose = getOption("verbose"),
+         mask.ok, exclude, include.only,
+         attach.required = missing(include.only))
 {
+    conf.ctrl <- getOption("conflicts.policy")
+    if (is.character(conf.ctrl))
+        conf.ctrl <-
+            switch(conf.ctrl,
+                   strict = list(error = TRUE, warn = FALSE),
+                   depends.ok = list(error = TRUE,
+                                     generics.ok = TRUE,
+                                     can.mask = c("base", "methods", "utils",
+                                                  "grDevices", "graphics",
+                                                  "stats"),
+                                     depends.ok = TRUE),
+                   warning(gettextf("unknown conflict policy: %s",
+                                    sQuote(conf.ctrl)),
+                           call. = FALSE, domain = NA))
+    if (! is.list(conf.ctrl))
+        conf.ctrl <- NULL
+    stopOnConflict <- isTRUE(conf.ctrl$error)
+
+    if (missing(warn.conflicts))
+        warn.conflicts <- if (isFALSE(conf.ctrl$warn)) FALSE else TRUE
+    if ((! missing(include.only)) && (! missing(exclude)))
+        stop(gettext("only one of 'include.only' and 'exclude' can be used"),
+             call. = FALSE, domain = NA)
+
     testRversion <- function(pkgInfo, pkgname, pkgpath)
     {
         if(is.null(built <- pkgInfo$Built))
@@ -55,16 +102,13 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
             for(dep in Rdeps)
                 if(length(dep) > 1L) {
                     target <- dep$version
-                    res <- if(is.character(target)) {
-                        do.call(dep$op, # these are both strings
-                                list(as.numeric(R.version[["svn rev"]]),
-                                     as.numeric(sub("^r", "", dep$version))))
-                    } else {
+                    res <-
                         do.call(dep$op,
-                                list(current, as.numeric_version(target)))
-##                        target <- as.numeric_version(dep$version)
-##                        eval(parse(text=paste("current", dep$op, "target")))
-                    }
+                           if(is.character(target)) # these are both strings
+                               list(as.numeric(R.version[["svn rev"]]),
+                                    as.numeric(sub("^r", "", target)))
+                           else
+                               list(current, as.numeric_version(target)))
                     if(!res)
                         stop(gettextf("This is R %s, package %s needs %s %s",
                                       current, sQuote(pkgname), dep$op, target),
@@ -82,11 +126,11 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
             ## allow mismatches if r_arch is in use, e.g.
             ## i386-gnu-linux vs x86-gnu-linux depending on
             ## build system.
-            if(!nzchar(r_arch) && length(grep("\\w", platform)) &&
-               !testPlatformEquivalence(platform, R.version$platform))
-                stop(gettextf("package %s was built for %s",
-                              sQuote(pkgname), platform),
-                     call. = FALSE, domain = NA)
+##             if(!nzchar(r_arch) && grepl("\\w", platform) &&
+##                !testPlatformEquivalence(platform, R.version$platform))
+##                 stop(gettextf("package %s was built for %s",
+##                               sQuote(pkgname), platform),
+##                      call. = FALSE, domain = NA)
         } else {  # Windows
             ## a check for 'mingw' suffices, since i386 and x86_64
             ## have DLLs in different places.  This allows binary packages
@@ -105,44 +149,6 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                  call. = FALSE, domain = NA)
     }
 
-    checkLicense <- function(pkg, pkgInfo, pkgPath)
-    {
-        L <- tools:::analyze_license(pkgInfo$DESCRIPTION["License"])
-        if(!L$is_empty && !L$is_verified) {
-            site_file <- path.expand(file.path(R.home("etc"), "licensed.site"))
-            if(file.exists(site_file) &&
-               pkg %in% readLines(site_file)) return()
-            personal_file <- path.expand("~/.R/licensed")
-            if(file.exists(personal_file)) {
-                agreed <- readLines(personal_file)
-                if(pkg %in% agreed) return()
-            } else agreed <- character()
-            if(!interactive())
-                stop(gettextf("package %s has a license that you need to accept in an interactive session", sQuote(pkg)), domain = NA)
-            lfiles <- file.path(pkgpath, c("LICENSE", "LICENCE"))
-            lfiles <- lfiles[file.exists(lfiles)]
-            if(length(lfiles)) {
-                message(gettextf("package %s has a license that you need to accept after viewing", sQuote(pkg)), domain = NA)
-                readline("press RETURN to view license")
-                encoding <- pkgInfo$DESCRIPTION["Encoding"]
-                if(is.na(encoding)) encoding <- ""
-                ## difR and EVER have a Windows' 'smart quote' LICEN[CS]E file
-                if(encoding == "latin1") encoding <- "cp1252"
-                file.show(lfiles[1L], encoding = encoding)
-            } else {
-                message(gettextf("package %s has a license that you need to accept:\naccording to the DESCRIPTION file it is", sQuote(pkg)), domain = NA)
-                message(pkgInfo$DESCRIPTION["License"], domain = NA)
-            }
-            choice <- menu(c("accept", "decline"),
-                           title = paste("License for", sQuote(pkg)))
-            if(choice != 1)
-                stop(gettextf("license for package %s not accepted",
-                              sQuote(package)), domain = NA, call. = FALSE)
-            dir.create(dirname(personal_file), showWarnings=FALSE)
-            writeLines(c(agreed, pkg), personal_file)
-        }
-    }
-
     checkNoGenerics <- function(env, pkg)
     {
         nenv <- env
@@ -153,11 +159,11 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
         else {
             ## A package will have created a generic
             ## only if it has created a formal method.
-            length(objects(env, pattern="^\\.__T", all.names=TRUE)) == 0L
+	    !any(startsWith(names(env), ".__T"))
         }
     }
 
-    ## FIXME: ./attach.R 's attach() has *very* similar checkConflicts(), keep in sync
+    ## NB: ./attach.R 's attach() has similar checkConflicts() [simpler at the end], keep in sync!
     checkConflicts <- function(package, pkgname, pkgpath, nogenerics, env)
     {
         dont.mind <- c("last.dump", "last.warning", ".Last.value",
@@ -165,29 +171,30 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                        ".packageName", ".noGenerics", ".required",
                        ".no_S3_generics", ".Depends", ".requireCachedGenerics")
         sp <- search()
-        lib.pos <- match(pkgname, sp)
+        lib.pos <- which(sp == pkgname)
         ## ignore generics not defined for the package
-        ob <- objects(lib.pos, all.names = TRUE)
+        ob <- names(as.environment(lib.pos))
         if(!nogenerics) {
             ##  Exclude generics that are consistent with implicit generic
             ## from another package.  A better test would be to move this
             ## down into the loop and test against specific other package name
             ## but subtle conflicts like that are likely to be found elsewhere
-	    these <- ob[substr(ob, 1L, 6L) == ".__T__"]
+	    these <- ob[startsWith(ob,".__T__")]
             gen  <- gsub(".__T__(.*):([^:]+)", "\\1", these)
             from <- gsub(".__T__(.*):([^:]+)", "\\2", these)
             gen <- gen[from != package]
             ob <- ob[!(ob %in% gen)]
         }
-        fst <- TRUE
-	ipos <- seq_along(sp)[-c(lib.pos,
-				 match(c("Autoloads", "CheckExEnv"), sp, 0L))]
+
+	ipos <- seq_along(sp)[-c(lib.pos, match(c("Autoloads", "CheckExEnv"), sp, 0L))]
+        cpos <- NULL
+        conflicts <- vector("list", 0)
         for (i in ipos) {
-            obj.same <- match(objects(i, all.names = TRUE), ob, nomatch = 0L)
-            if (any(obj.same > 0)) {
+            obj.same <- match(names(as.environment(i)), ob, nomatch = 0L)
+            if (any(obj.same > 0L)) {
                 same <- ob[obj.same]
                 same <- same[!(same %in% dont.mind)]
-                Classobjs <- grep("^\\.__", same)
+                Classobjs <- which(startsWith(same,".__"))
                 if(length(Classobjs)) same <- same[-Classobjs]
                 ## report only objects which are both functions or
                 ## both non-functions.
@@ -207,14 +214,58 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
 		if(length(same) && identical(sp[i], "package:base"))
 		    same <- same[not.Ident(same, ignore.environment = TRUE)]
                 if(length(same)) {
-                    if (fst) {
-                        fst <- FALSE
-                        packageStartupMessage(gettextf("\nAttaching package: %s\n",
-                                                       sQuote(package)),
-                                              domain = NA)
+                    conflicts[[sp[i]]] <- same
+                    cpos[sp[i]] <- i
+                }
+            }
+        }
+        if (length(conflicts)) {
+            if (stopOnConflict) {
+                emsg <- ""
+                pkg <- names(conflicts)
+                notOK <- vector("list", 0)
+                for (i in seq_along(conflicts)) {
+                    pkgname <- sub("^package:", "", pkg[i])
+                    if (pkgname %in% canMaskEnv$canMask)
+                        next
+                    same <- conflicts[[i]]
+                    if (is.list(mask.ok))
+                        myMaskOK <- mask.ok[[pkgname]]
+                    else myMaskOK <- mask.ok
+
+                    ## adjust 'same' for conflict resolution specifications
+                    if (isTRUE(myMaskOK))
+                        same <- NULL
+                    else if (is.character(myMaskOK))
+                        same <- setdiff(same, myMaskOK)
+
+                    if (length(same)) {
+                        notOK[[pkg[i]]] <- same
+                        msg <- .maskedMsg(sort(same), pkg = sQuote(pkg[i]),
+                                          by = cpos[i] < lib.pos)
+                        emsg <- paste(emsg, msg, sep = "\n")
                     }
-		    msg <- .maskedMsg(same, pkg = sQuote(sp[i]), by = i < lib.pos)
-		    packageStartupMessage(msg, domain = NA)
+                }
+                if (length(notOK)) {
+                    msg <- gettextf("Conflicts attaching package %s:\n%s",
+                                    sQuote(package),
+                                    emsg)
+                    stop(errorCondition(msg,
+                                        package = package,
+                                        conflicts = conflicts,
+                                        class = "packageConflictError"))
+                }
+            }
+            if (warn.conflicts) {
+                ## Use separate messages to preserve previous behavior.
+                packageStartupMessage(gettextf("\nAttaching package: %s\n",
+                                               sQuote(package)), domain = NA)
+                pkg <- names(conflicts)
+                for (i in seq_along(conflicts)) {
+                    msg <- .maskedMsg(sort(conflicts[[i]]),
+                                      pkg = sQuote(pkg[i]),
+                                      by = cpos[i] < lib.pos)
+                    packageStartupMessage(msg, domain = NA)
                 }
             }
         }
@@ -225,7 +276,7 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
     if(!missing(package)) {
         if (is.null(lib.loc)) lib.loc <- .libPaths()
         ## remove any non-existent directories
-        lib.loc <- lib.loc[file.info(lib.loc)$isdir %in% TRUE]
+        lib.loc <- lib.loc[dir.exists(lib.loc)]
 
 	if(!character.only)
 	    package <- as.character(substitute(package))
@@ -234,7 +285,7 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
         if(is.na(package) || (package == ""))
             stop("invalid package name")
 
-	pkgname <- paste("package", package, sep = ":")
+	pkgname <- paste0("package:", package)
 	newpackage <- is.na(match(pkgname, search()))
 	if(newpackage) {
             ## Check for the methods package before attaching this
@@ -243,9 +294,14 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
             ## The methods package caches all other pkgs when it is
             ## attached.
 
-            pkgpath <- find.package(package, lib.loc, quiet = TRUE,
+            ## Too extreme (unfortunately; warning too often):
+	    ## pkgpath <- find.package(package, lib.loc, quiet = TRUE, verbose = !quietly)
+	    ##   'verbose' here means to warn about packages found more than once
+	    pkgpath <- find.package(package, lib.loc, quiet = TRUE,
                                     verbose = verbose)
             if(length(pkgpath) == 0L) {
+                if(length(lib.loc) && !logical.return)
+                    stop(packageNotFoundError(package, lib.loc, sys.call()))
                 txt <- if(length(lib.loc))
                     gettextf("there is no package called %s", sQuote(package))
                 else
@@ -263,58 +319,95 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                               sQuote(package)), domain = NA)
             pkgInfo <- readRDS(pfile)
             testRversion(pkgInfo, package, pkgpath)
-            ## avoid any bootstrapping issues by these exemptions
-            if(!package %in% c("datasets", "grDevices", "graphics", "methods",
-                               "splines", "stats", "stats4", "tcltk", "tools",
-                               "utils") &&
-               isTRUE(getOption("checkPackageLicense", FALSE)))
-                checkLicense(package, pkgInfo, pkgpath)
 
+            ## The ABI compatibility check is now in loadNamespace
+            ## The licence check is now in loadNamespace
             ## The check for inconsistent naming is now in find.package
 
             if(is.character(pos)) {
                 npos <- match(pos, search())
                 if(is.na(npos)) {
-                    warning(gettextf("%s not found on search path, using pos = 2", sQuote(pos)), domain = NA)
+                    warning(gettextf("%s not found on search path, using pos = 2",
+                                     sQuote(pos)), domain = NA)
                     pos <- 2
                 } else pos <- npos
             }
-            .getRequiredPackages2(pkgInfo, quietly = quietly)
+
             deps <- unique(names(pkgInfo$Depends))
+            depsOK <- isTRUE(conf.ctrl$depends.ok)
+            if (depsOK) {
+                canMaskEnv <- dynGet("__library_can_mask__", NULL)
+                if (is.null(canMaskEnv)) {
+                    canMaskEnv <- new.env()
+                    canMaskEnv$canMask <- union("base", conf.ctrl$can.mask)
+                    "__library_can_mask__" <- canMaskEnv
+                }
+                canMaskEnv$canMask <- unique(c(package, deps,
+                                               canMaskEnv$canMask))
+            }
+            else canMaskEnv <- NULL
+
+            if (attach.required)
+                .getRequiredPackages2(pkgInfo, quietly = quietly)
+
+            cr <- conflictRules(package)
+            if (missing(mask.ok))
+                mask.ok <- cr$mask.ok
+            if (missing(exclude))
+                exclude <- cr$exclude
 
             ## If the namespace mechanism is available and the package
             ## has a namespace, then the namespace loading mechanism
             ## takes over.
             if (packageHasNamespace(package, which.lib.loc)) {
-                if (package %in% loadedNamespaces()) {
-                    # Already loaded.  Does the version match?
+		if (isNamespaceLoaded(package)) {
+                    ## Already loaded.  Does the version match?
                     newversion <- as.numeric_version(pkgInfo$DESCRIPTION["Version"])
                     oldversion <- as.numeric_version(getNamespaceVersion(package))
                     if (newversion != oldversion) {
-                    	# No, so try to unload the previous one
-                    	res <- try(unloadNamespace(package))
-                    	if (inherits(res, "try-error"))
-                    	    stop(dQuote(package), " version ", oldversion, " cannot be unloaded.")
+                    	## No, so try to unload the previous one
+			tryCatch(unloadNamespace(package),
+				 error = function(e) {
+				     P <- if(!is.null(cc <- conditionCall(e)))
+					      paste("Error in", deparse(cc)[1L], ": ")
+					  else "Error : "
+				     stop(gettextf("Package %s version %s cannot be unloaded:\n %s",
+						   sQuote(package), oldversion,
+						   paste0(P, conditionMessage(e),"\n")),
+					  domain=NA)})
                     }
                 }
-                tt <- try({
-                    ns <- loadNamespace(package, c(which.lib.loc, lib.loc))
-                    env <- attachNamespace(ns, pos = pos, deps)
-                })
-                if (inherits(tt, "try-error"))
-                    if (logical.return)
-                        return(FALSE)
-                    else stop(gettextf("package or namespace load failed for %s",
-                                       sQuote(package)),
-                              call. = FALSE, domain = NA)
-                else {
+		tt <- tryCatch({
+                    attr(package, "LibPath") <- which.lib.loc
+                    ns <- loadNamespace(package, lib.loc)
+                    env <- attachNamespace(ns, pos = pos, deps,
+                                           exclude, include.only)
+		}, error = function(e) {
+		    P <- if(!is.null(cc <- conditionCall(e)))
+			     paste(" in", deparse(cc)[1L]) else ""
+		    msg <- gettextf("package or namespace load failed for %s%s:\n %s",
+				    sQuote(package), P, conditionMessage(e))
+		    if(logical.return)
+			message(paste("Error:", msg), domain = NA) # returns NULL
+		    else stop(msg, call. = FALSE, domain = NA)
+		})
+		if(logical.return && is.null(tt))
+		    return(FALSE)
+
+                attr(package, "LibPath") <- NULL
+                {
                     on.exit(detach(pos = pos))
                     ## If there are S4 generics then the package should
                     ## depend on methods
                     nogenerics <-
                         !.isMethodsDispatchOn() || checkNoGenerics(env, package)
-                    if(warn.conflicts && # never will with a namespace
-                       !exists(".conflicts.OK", envir = env, inherits = FALSE))
+                    if (isFALSE(conf.ctrl$generics.ok) ||
+                        (stopOnConflict && ! isTRUE(conf.ctrl$generics.ok)))
+                        nogenerics <- TRUE ## no silent masking for genrics
+                    if(stopOnConflict ||
+                       (warn.conflicts && # never will with a namespace
+                        !exists(".conflicts.OK", envir = env,
+                                inherits = FALSE)))
                         checkConflicts(package, pkgname, pkgpath,
                                        nogenerics, ns)
                     on.exit()
@@ -324,8 +417,8 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                         return(invisible(.packages()))
                 }
             } else
-            stop(gettextf("package %s does not have a namespace and should be re-installed",
-                          sQuote(package)), domain = NA)
+		stop(gettextf("package %s does not have a namespace and should be re-installed",
+			      sQuote(package)), domain = NA)
 	}
 	if (verbose && !newpackage)
             warning(gettextf("package %s already present in search()",
@@ -352,10 +445,12 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                     if(!inherits(tmp, "try-error"))
                         txt <- tmp
                     else
-                        warning("'DESCRIPTION' has an 'Encoding' field and re-encoding is not possible", call.=FALSE)
+                        warning("'DESCRIPTION' has an 'Encoding' field and re-encoding is not possible",
+                                call. = FALSE)
                 }
                 nm <- paste0(names(txt), ":")
-                formatDL(nm, txt, indent = max(nchar(nm, "w")) + 3)
+                ## indent might be excessive for long field names.
+                formatDL(nm, txt, indent = max(nchar(nm, "w")) + 3L)
             } else if(basename(f) %in% "vignette.rds") {
                 txt <- readRDS(f)
                 ## New-style vignette indices are data frames with more
@@ -369,7 +464,7 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                                         txt$File)),
                           paste(txt$Title,
                                 paste0(rep.int("(source", NROW(txt)),
-                                       ifelse(txt$PDF != "",
+                                       ifelse(nzchar(txt$PDF),
                                               ", pdf",
                                               ""),
                                        ")")))
@@ -397,7 +492,7 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
                 ## 'package.rds' but we have not checked.
                 file <- system.file("Meta", "package.rds", package = i,
                                     lib.loc = lib)
-                title <- if(file != "") {
+                title <- if(nzchar(file)) {
                     txt <- readRDS(file)
                     if(is.list(txt)) txt <- txt$DESCRIPTION
                     ## we may need to re-encode here.
@@ -436,7 +531,7 @@ function(package, help, pos = 2, lib.loc = NULL, character.only = FALSE,
     if (logical.return)
 	TRUE
     else invisible(.packages())
-}
+} ## {library}
 
 format.libraryIQR <-
 function(x, ...)
@@ -531,8 +626,9 @@ function(chname, package, lib.loc, verbose = getOption("verbose"),
     }
     if(verbose)
         message(gettextf("now dyn.load(\"%s\") ...", file), domain = NA)
-    dll <- if("DLLpath" %in% names(list(...))) dyn.load(file, ...)
-    else dyn.load(file, DLLpath = DLLpath, ...)
+    dll <- if("DLLpath" %in% ...names())
+                dyn.load(file, ...)
+           else dyn.load(file, DLLpath = DLLpath, ...)
     .dynLibs(c(dll_list, list(dll)))
     invisible(dll)
 }
@@ -580,12 +676,13 @@ function(chname, libpath, verbose = getOption("verbose"),
 }
 
 require <-
-function(package, lib.loc = NULL, quietly = FALSE, warn.conflicts = TRUE,
-         character.only = FALSE)
+function(package, lib.loc = NULL, quietly = FALSE, warn.conflicts,
+         character.only = FALSE, mask.ok, exclude, include.only,
+         attach.required = missing(include.only))
 {
     if(!character.only)
         package <- as.character(substitute(package)) # allowing "require(eda)"
-    loaded <- paste("package", package, sep = ":") %in% search()
+    loaded <- paste0("package:", package) %in% search()
 
     if (!loaded) {
 	if (!quietly)
@@ -595,7 +692,11 @@ function(package, lib.loc = NULL, quietly = FALSE, warn.conflicts = TRUE,
                                   character.only = TRUE,
                                   logical.return = TRUE,
                                   warn.conflicts = warn.conflicts,
-				  quietly = quietly),
+				  quietly = quietly,
+                                  mask.ok = mask.ok,
+                                  exclude = exclude,
+                                  include.only = include.only,
+                                  attach.required = attach.required),
                           error = function(e) e)
         if (inherits(value, "error")) {
             if (!quietly) {
@@ -626,7 +727,7 @@ function(all.available = FALSE, lib.loc = NULL)
         return(unique(ans))
     } ## else
     s <- search()
-    return(invisible(substring(s[substr(s, 1L, 8L) == "package:"], 9)))
+    invisible(.rmpkg(s[startsWith(s, "package:")]))
 }
 
 path.package <-
@@ -638,7 +739,7 @@ function(package = NULL, quiet = FALSE)
     searchpaths <-
         lapply(seq_along(s), function(i) attr(as.environment(i), "path"))
     searchpaths[[length(s)]] <- system.file()
-    pkgs <- paste("package", package, sep = ":")
+    pkgs <- paste0("package:", package)
     pos <- match(pkgs, s)
     if(any(m <- is.na(pos))) {
         if(!quiet) {
@@ -672,37 +773,23 @@ function(package = NULL, lib.loc = NULL, quiet = FALSE,
     if(length(package) == 1L  &&
        package %in% c("base", "tools", "utils", "grDevices", "graphics",
                       "stats", "datasets", "methods", "grid", "parallel",
-                      "splines", "stats4", "tcltk"))
+                      "splines", "stats4", "tcltk", "compiler"))
         return(file.path(.Library, package))
 
-    use_loaded <- FALSE
     if(is.null(package)) package <- .packages()
-    if(is.null(lib.loc)) {
-        use_loaded <- TRUE
-        lib.loc <- .libPaths()
-    }
-
     if(!length(package)) return(character())
+    if(use_loaded <- is.null(lib.loc))
+	lib.loc <- .libPaths()
 
     bad <- character()
     out <- character()
 
     for(pkg in package) {
-        paths <- character()
-        for(lib in lib.loc) {
-            dirs <- list.files(lib,
-                               pattern = paste0("^", pkg, "$"),
-                               full.names = TRUE)
-            ## Note that we cannot use tools::file_test() here, as
-            ## cyclic namespace dependencies are not supported.  Argh.
-            paths <- c(paths,
-                       dirs[file.info(dirs)$isdir &
-                            file.exists(file.path(dirs,
-                                                  "DESCRIPTION"))])
-        }
-        if(use_loaded && pkg %in% loadedNamespaces()) {
-            dir <- if (pkg == "base") system.file()
-            else getNamespaceInfo(pkg, "path")
+	paths <- file.path(lib.loc, pkg)
+	paths <- paths[ file.exists(file.path(paths, "DESCRIPTION")) ]
+	if(use_loaded && isNamespaceLoaded(pkg)) {
+	    dir <- if (pkg == "base") system.file()
+		   else .getNamespaceInfo(asNamespace(pkg), "path")
             paths <- c(dir, paths)
         }
         ## trapdoor for tools:::setRlibs
@@ -719,17 +806,22 @@ function(package = NULL, lib.loc = NULL, quiet = FALSE,
                 ## Note that this is sometimes used for source
                 ## packages, e.g. by promptPackage from package.skeleton
                 pfile <- file.path(p, "Meta", "package.rds")
-                info <- if(file.exists(pfile))
+                info <- if(file.exists(pfile)) {
                     ## this must have these fields to get installed
-                    readRDS(pfile)$DESCRIPTION[c("Package", "Version")]
-                else {
+                    tryCatch(readRDS(pfile)$DESCRIPTION[c("Package",
+                                                          "Version")],
+                             error = function(e)
+                                 c(Package = NA_character_,
+                                   Version = NA_character_))
+                } else {
                     info <- tryCatch(read.dcf(file.path(p, "DESCRIPTION"),
                                               c("Package", "Version"))[1, ],
                                      error = identity)
                     if(inherits(info, "error")
                        || (length(info) != 2L)
                        || anyNA(info))
-                        c(Package = NA, Version = NA) # need dimnames below
+                        c(Package = NA_character_,
+                          Version = NA_character_) # need dimnames below
                     else
                         info
                 }
@@ -747,27 +839,19 @@ function(package = NULL, lib.loc = NULL, quiet = FALSE,
         }
         if(length(paths) > 1L) {
             ## If a package was found more than once ...
+	    if(verbose)
+		warning(gettextf("package %s found more than once, using the first from\n  %s",
+				 sQuote(pkg),
+				 paste(dQuote(paths), collapse=",\n  ")),
+			domain = NA)
             paths <- paths[1L]
-            if(verbose)
-                warning(gettextf("package %s found more than once,\nusing the one found in %s",
-                                 sQuote(pkg), sQuote(paths)), domain = NA)
         }
         out <- c(out, paths)
     }
 
     if(!quiet && length(bad)) {
-        if(length(out) == 0L) {
-            if(length(bad) == 1L) {
-                stop(gettextf("there is no package called %s", sQuote(pkg)),
-                     domain = NA)
-            } else {
-                stop(ngettext(length(bad),
-                              "there is no package called",
-                              "there are no packages called"), " ",
-                     paste(sQuote(bad), collapse = ", "), domain = NA)
-
-            }
-        }
+        if(length(out) == 0L)
+            stop(packageNotFoundError(bad, lib.loc, sys.call()))
         for(pkg in bad)
             warning(gettextf("there is no package called %s", sQuote(pkg)),
                     domain = NA)
@@ -776,6 +860,19 @@ function(package = NULL, lib.loc = NULL, quiet = FALSE,
     out
 }
 
+packageNotFoundError <-
+function(package, lib.loc, call = NULL) {
+    if(length(package) == 1L)
+        msg <- gettextf("there is no package called %s", sQuote(package))
+    else
+        msg <- paste0(ngettext(length(package),
+                               "there is no package called",
+                               "there are no packages called"), " ",
+                      paste(sQuote(package), collapse = ", "))
+    errorCondition(msg, package = package, lib.loc = lib.loc, call = call,
+                   class = "packageNotFoundError")
+}
+
 format.packageInfo <-
 function(x, ...)
 {
@@ -823,61 +920,61 @@ function(file="DESCRIPTION", lib.loc = NULL, quietly = FALSE, useImports = FALSE
 .getRequiredPackages2 <-
 function(pkgInfo, quietly = FALSE, lib.loc = NULL, useImports = FALSE)
 {
+### FIXME: utils::packageVersion() should be pushed up here instead
+    .findVersion <- function(pkg, lib.loc) {
+        pfile <- system.file("Meta", "package.rds",
+                             package = pkg, lib.loc = lib.loc)
+        if (nzchar(pfile))
+            as.numeric_version(readRDS(pfile)$DESCRIPTION["Version"])
+        ## else NULL
+    }
     pkgs <- unique(names(pkgInfo$Depends))
-    if (length(pkgs)) {
-        pkgname <- pkgInfo$DESCRIPTION["Package"]
-        for(pkg in pkgs) {
-            ## several packages 'Depends' on base!
-            if (pkg == "base") next
-            ## allow for multiple occurrences
-            zs <- pkgInfo$Depends[names(pkgInfo$Depends) == pkg]
-            have_vers <- any(vapply(zs, length, 1L) > 1L)
-            if ( !paste("package", pkg, sep = ":") %in% search() ) {
-                if (have_vers) {
-                    pfile <- system.file("Meta", "package.rds",
-                                         package = pkg, lib.loc = lib.loc)
-                    if(!nzchar(pfile))
-                        stop(gettextf("package %s required by %s could not be found",
-                                      sQuote(pkg), sQuote(pkgname)),
-                             call. = FALSE, domain = NA)
-                    current <- readRDS(pfile)$DESCRIPTION["Version"]
-                    for(z in zs)
-                        if(length(z) > 1L) {
-                            target <- as.numeric_version(z$version)
-                            if (!do.call(z$op, list(as.numeric_version(current), target)))
-##                            if (!eval(parse(text=paste("current", z$op, "target"))))
-                                stop(gettextf("package %s %s was found, but %s %s is required by %s",
-                                              sQuote(pkg), current, z$op,
-                                              target, sQuote(pkgname)),
-                                     call. = FALSE, domain = NA)
-                        }
+    pkgname <- pkgInfo$DESCRIPTION["Package"]
+    for(pkg in setdiff(pkgs, "base")) {
+        ## allow for multiple occurrences
+        depends <- pkgInfo$Depends[names(pkgInfo$Depends) == pkg]
+        attached <- paste0("package:", pkg) %in% search()
+        current <- .findVersion(pkg, lib.loc)
+        if(is.null(current))
+            stop(gettextf("package %s required by %s could not be found",
+                          sQuote(pkg), sQuote(pkgname)),
+                 call. = FALSE, domain = NA)
+        have_vers <- lengths(depends) > 1L
+        for(dep in depends[have_vers]) {
+            target <- as.numeric_version(dep$version)
+            sufficient <- do.call(dep$op, list(current, target))
+            if (!sufficient) {
+                if (is.null(lib.loc))
+                    lib.loc <- .libPaths()
+		allV <- lapply(lib.loc, .findVersion, pkg=pkg)
+		versions <- do.call(c, allV[iV <- which(!vapply(allV, is.null, NA))])
+                sufficient <- vapply(versions, dep$op, logical(1L), target)
+                if (any(sufficient)) {
+                    warning(gettextf("version %s of %s masked by %s in %s",
+                                     versions[which(sufficient)[1L]],
+                                     sQuote(pkg),
+                                     current,
+				     lib.loc[iV[!sufficient][1L]]),
+                            call. = FALSE, domain = NA)
                 }
+		msg <- if (attached)
+			   "package %s %s is loaded, but %s %s is required by %s"
+		       else
+			   "package %s %s was found, but %s %s is required by %s"
+                stop(gettextf(msg, sQuote(pkg), current, dep$op,
+                              target, sQuote(pkgname)),
+                     call. = FALSE, domain = NA)
+            }
+        }
 
-                if (!quietly)
-                    packageStartupMessage(gettextf("Loading required package: %s",
-                                     pkg), domain = NA)
-                library(pkg, character.only = TRUE, logical.return = TRUE,
-                        lib.loc = lib.loc) ||
+        if (!attached) {
+            if (!quietly)
+                packageStartupMessage(gettextf("Loading required package: %s",
+                                               pkg), domain = NA)
+            library(pkg, character.only = TRUE, logical.return = TRUE,
+                    lib.loc = lib.loc, quietly = quietly) ||
                 stop(gettextf("package %s could not be loaded", sQuote(pkg)),
                      call. = FALSE, domain = NA)
-            } else {
-                ## check the required version number, if any
-                if (have_vers) {
-                    pfile <- system.file("Meta", "package.rds",
-                                         package = pkg, lib.loc = lib.loc)
-                    current <- readRDS(pfile)$DESCRIPTION["Version"]
-                    for(z in zs)
-                        if (length(z) > 1L) {
-                            target <- as.numeric_version(z$version)
-                            if (!do.call(z$op, list(as.numeric_version(current), target)))
-##                            if (!eval(parse(text=paste("current", z$op, "target"))))
-                                stop(gettextf("package %s %s is loaded, but %s %s is required by %s",
-                                              sQuote(pkg), current, z$op,
-                                              target, sQuote(pkgname)),
-                                     call. = FALSE, domain = NA)
-                        }
-                }
-            }
         }
     }
     if(useImports) {
@@ -906,5 +1003,5 @@ function(x)
     ## %o => os
     x <- expand(x, "o", R.version$os)
 
-    gsub("%%", "%", x)
+    gsub("%%", "%", x, fixed=TRUE)
 }
