@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
+ *  Copyright (C) 1997--2019  The R Core Team
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2015  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
 
@@ -27,9 +27,10 @@
 #include <Internal.h>
 #include <R_ext/Random.h>
 
-/* Normal generator is not actually set here but in nmath/snorm.c */
+/* Normal generator is not actually set here but in ../nmath/snorm.c */
 #define RNG_DEFAULT MERSENNE_TWISTER
 #define N01_DEFAULT INVERSION
+#define Sample_DEFAULT REJECTION
 
 
 #include <R_ext/Rdynload.h>
@@ -42,14 +43,18 @@ UnifInitFun User_unif_init = NULL; /* some picky compilers */
 
 DL_FUNC  User_norm_fun = NULL; /* also in ../nmath/snorm.c */
 
-
+#include "nmath2.h"
 static RNGtype RNG_kind = RNG_DEFAULT;
-extern N01type N01_kind; /* from ../nmath/snorm.c */
+//extern N01type N01_kind; /* from ../nmath/snorm.c */
+//extern double BM_norm_keep; /* ../nmath/snorm.c */
+static Sampletype Sample_kind = REJECTION;
 
 /* typedef unsigned int Int32; in Random.h */
 
 /* .Random.seed == (RNGkind, i_seed[0],i_seed[1],..,i_seed[n_seed-1])
  * or           == (RNGkind) or missing  [--> Randomize]
+ * where  RNGkind :=  RNG_kind  +  100 * N01_kind  +  10000 * Sample_kind   
+ * currently in  outer(outer(0:7, 100*(0:5), "+"), 10000*(0:1), "+")
  */
 
 typedef struct {
@@ -61,7 +66,7 @@ typedef struct {
 } RNGTAB;
 
 
-static Int32 dummy[625];
+static Int32 dummy[628]; // allow for optimizing compilers to read over bound
 static
 RNGTAB RNG_Table[] =
 {
@@ -261,8 +266,6 @@ static void FixupSeeds(RNGtype RNG_kind, int initial)
     }
 }
 
-extern double BM_norm_keep; /* ../nmath/snorm.c */
-
 static void RNG_Init(RNGtype kind, Int32 seed)
 {
     int j;
@@ -332,23 +335,21 @@ static SEXP GetSeedsFromVar(void)
     return seeds;
 }
 
-unsigned int TimeToSeed(void); // times.c
-
 static void Randomize(RNGtype kind)
 {
 /* Only called by  GetRNGstate() when there is no .Random.seed */
     RNG_Init(kind, TimeToSeed());
 }
 
-static void GetRNGkind(SEXP seeds)
+static Rboolean GetRNGkind(SEXP seeds)
 {
-    /* Load RNG_kind, N01_kind from .Random.seed if present */
+    /* Load RNG_kind, N01_kind Sample_kind from .Random.seed if present */
     int tmp, *is;
-    RNGtype newRNG; N01type newN01;
+    RNGtype newRNG; N01type newN01; Sampletype newSample;
 
     if (isNull(seeds))
 	seeds = GetSeedsFromVar();
-    if (seeds == R_UnboundValue) return;
+    if (seeds == R_UnboundValue) return TRUE;
     if (!isInteger(seeds)) {
 	if (seeds == R_MissingArg) /* How can this happen? */
 	    error(_("'.Random.seed' is a missing argument with no default"));
@@ -358,14 +359,15 @@ static void GetRNGkind(SEXP seeds)
     }
     is = INTEGER(seeds);
     tmp = is[0];
-    /* avoid overflow here: max current value is 705 */
-    if (tmp == NA_INTEGER || tmp < 0 || tmp > 1000) {
+    /* avoid overflow here: max current value is 10705 */
+    if (tmp == NA_INTEGER || tmp < 0 || tmp > 11000) {
 	warning(_("'.Random.seed[1]' is not a valid integer, so ignored"));
 	goto invalid;
     }
     newRNG = (RNGtype) (tmp % 100);
-    newN01 = (N01type) (tmp / 100);
-    if (newN01 > KINDERMAN_RAMAGE) {
+    newN01 = (N01type) (tmp % 10000 / 100);
+    newSample = (Sampletype) (tmp / 10000);
+    if (newN01 > KINDERMAN_RAMAGE || newSample > REJECTION) {
 	warning(_("'.Random.seed[1]' is not a valid Normal type, so ignored"));
 	goto invalid;
     }
@@ -388,12 +390,14 @@ static void GetRNGkind(SEXP seeds)
 	warning(_("'.Random.seed[1]' is not a valid RNG kind so ignored"));
 	goto invalid;
     }
-    RNG_kind = newRNG; N01_kind = newN01;
-    return;
+    RNG_kind = newRNG; N01_kind = newN01; Sample_kind = newSample;
+    return FALSE;
 invalid:
-    RNG_kind = RNG_DEFAULT; N01_kind = N01_DEFAULT;
+    RNG_kind = RNG_DEFAULT; N01_kind = N01_DEFAULT; Sample_kind = Sample_DEFAULT;
+
     Randomize(RNG_kind);
-    return;
+    PutRNGstate(); // write out to .Random.seed
+    return TRUE;
 }
 
 
@@ -403,12 +407,12 @@ void GetRNGstate()
     int len_seed;
     SEXP seeds;
 
-    /* look only in the workspace */
     seeds = GetSeedsFromVar();
     if (seeds == R_UnboundValue) {
 	Randomize(RNG_kind);
     } else {
-	GetRNGkind(seeds);
+	/* this might re-set the generator */
+	if(GetRNGkind(seeds)) return;
 	len_seed = RNG_Table[RNG_kind].n_seed;
 	/* Not sure whether this test is needed: wrong for USER_UNIF */
 	if(LENGTH(seeds) > 1 && LENGTH(seeds) < len_seed + 1)
@@ -430,7 +434,7 @@ void PutRNGstate()
     int len_seed, j;
     SEXP seeds;
 
-    if (RNG_kind > LECUYER_CMRG || N01_kind > KINDERMAN_RAMAGE) {
+    if (RNG_kind > LECUYER_CMRG || N01_kind > KINDERMAN_RAMAGE || Sample_kind > REJECTION) {
 	warning("Internal .Random.seed is corrupt: not saving");
 	return;
     }
@@ -439,7 +443,7 @@ void PutRNGstate()
 
     PROTECT(seeds = allocVector(INTSXP, len_seed + 1));
 
-    INTEGER(seeds)[0] = RNG_kind + 100 * N01_kind;
+    INTEGER(seeds)[0] = RNG_kind + 100 * N01_kind + 10000 * Sample_kind;
     for(j = 0; j < len_seed; j++)
 	INTEGER(seeds)[j+1] = RNG_Table[RNG_kind].i_seed[j];
 
@@ -497,20 +501,34 @@ static void Norm_kind(N01type kind)
     PutRNGstate();
 }
 
+static void Samp_kind(Sampletype kind)
+{
+    /* Sampletype is an enumeration type, so this will probably get
+       mapped to an unsigned integer type. */
+    if (kind == (Sampletype)-1) kind = Sample_DEFAULT;
+    if (kind > REJECTION)
+        error(_("invalid sample type in 'RNGkind'"));
+    GetRNGstate(); /* might not be initialized */
+    Sample_kind = kind;
+    PutRNGstate();
+}
+
 
 /*------ .Internal interface ------------------------*/
 
 SEXP attribute_hidden do_RNGkind (SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP ans, rng, norm;
+    SEXP ans, rng, norm, sample;
 
     checkArity(op,args);
     GetRNGstate(); /* might not be initialized */
-    PROTECT(ans = allocVector(INTSXP, 2));
+    PROTECT(ans = allocVector(INTSXP, 3));
     INTEGER(ans)[0] = RNG_kind;
     INTEGER(ans)[1] = N01_kind;
+    INTEGER(ans)[2] = Sample_kind;
     rng = CAR(args);
     norm = CADR(args);
+    sample = CADDR(args);
     GetRNGkind(R_NilValue); /* pull from .Random.seed if present */
     if(!isNull(rng)) { /* set a new RNG kind */
 	RNGkind((RNGtype) asInteger(rng));
@@ -518,6 +536,9 @@ SEXP attribute_hidden do_RNGkind (SEXP call, SEXP op, SEXP args, SEXP env)
     if(!isNull(norm)) { /* set a new normal kind */
 	Norm_kind((N01type) asInteger(norm));
     }
+    if(!isNull(sample)) { /* set a new sample kind */
+	Samp_kind((Sampletype) asInteger(sample));
+    }
     UNPROTECT(1);
     return ans;
 }
@@ -525,7 +546,7 @@ SEXP attribute_hidden do_RNGkind (SEXP call, SEXP op, SEXP args, SEXP env)
 
 SEXP attribute_hidden do_setseed (SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP skind, nkind;
+    SEXP skind, nkind, sampkind;
     int seed;
 
     checkArity(op, args);
@@ -536,10 +557,12 @@ SEXP attribute_hidden do_setseed (SEXP call, SEXP op, SEXP args, SEXP env)
     } else seed = TimeToSeed();
     skind = CADR(args);
     nkind = CADDR(args);
-    GetRNGkind(R_NilValue); /* pull RNG_kind, N01_kind from 
+    sampkind = CADDDR(args);
+    GetRNGkind(R_NilValue); /* pull RNG_kind, N01_kind from
 			       .Random.seed if present */
     if (!isNull(skind)) RNGkind((RNGtype) asInteger(skind));
     if (!isNull(nkind)) Norm_kind((N01type) asInteger(nkind));
+    if(!isNull(sampkind)) Samp_kind((Sampletype) asInteger(sampkind));
     RNG_Init(RNG_kind, (Int32) seed); /* zaps BM history */
     PutRNGstate();
     return R_NilValue;
@@ -563,6 +586,14 @@ void seed_out(long *ignored)
 
 /* ===================  Mersenne Twister ========================== */
 /* From http://www.math.keio.ac.jp/~matumoto/emt.html */
+/* New URL (accessed 2018-11-08):
+   http://www.math.sci.hiroshima-u.ac.jp/~m-mat/eindex.html
+
+   The initialization method in the 1998 code and paper had a minor
+   issue that was addressed with new initialization approaches in an
+   update in 2002.  R has always used a different initialization
+   approach and is not affected by that issue.
+*/
 
 /* A C-program for MT19937: Real number version([0,1)-interval)
    (1999/10/28)
@@ -674,9 +705,16 @@ static double MT_genrand(void)
 */
 
 
-/* This define may give a warning in clang, but is needed to comply
+/* This define may give a warning with clang, but is needed to comply
    with the prohibition on changing the code. */
+#ifdef __clang__
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wkeyword-macro"
+#endif
 #define long Int32
+#ifdef __clang__
+#pragma clang diagnostic pop
+#endif
 
 #define ran_arr_buf       R_KT_ran_arr_buf
 #define ran_arr_cycle     R_KT_ran_arr_cycle
@@ -770,13 +808,71 @@ static Int32 KT_next(void)
 static void RNG_Init_R_KT(Int32 seed)
 {
     SEXP fun, sseed, call, ans;
-    fun = findVar1(install(".TAOCP1997init"), R_BaseEnv, CLOSXP, FALSE);
+    PROTECT(fun = findVar1(install(".TAOCP1997init"), R_BaseEnv, CLOSXP, FALSE));
     if(fun == R_UnboundValue)
 	error("function '.TAOCP1997init' is missing");
     PROTECT(sseed = ScalarInteger((int)(seed % 1073741821)));
     PROTECT(call = lang2(fun, sseed));
     ans = eval(call, R_GlobalEnv);
     memcpy(dummy, INTEGER(ans), 100*sizeof(int));
-    UNPROTECT(2);
+    UNPROTECT(3);
     KT_pos = 100;
 }
+
+/* Our PRNGs have at most 32 bit of precision. All generators except
+   Knuth-TAOCP, Knuth-TAOCP-2002, and possibly the user-supplied ones
+   have 31 or 32 bits of precision; the others are assumed to
+   have at least 25. */
+static R_INLINE double ru()
+{
+    double U = 33554432.0;
+    return (floor(U*unif_rand()) + unif_rand())/U;
+}
+
+static double R_unif_index_0(double dn)
+{
+    double cut = INT_MAX;
+
+    switch(RNG_kind) {
+    case KNUTH_TAOCP:
+    case USER_UNIF:
+    case KNUTH_TAOCP2:
+	cut = 33554431.0; /* 2^25 - 1 */
+ 	break;
+    default:
+ 	break;
+    }
+
+    double u = dn > cut ? ru() : unif_rand();
+    return floor(dn * u);
+}
+
+//generate a random non-negative integer < 2 ^ bits in 16 bit chunks
+static double rbits(int bits)
+{
+    int_least64_t v = 0;
+    for (int n = 0; n <= bits; n += 16) {
+	int v1 = (int) floor(unif_rand() * 65536);
+	v = 65536 * v + v1;
+    }
+    const int_least64_t one64 = 1L;
+    // mask out the bits in the result that are not needed
+    return (double) (v & ((one64 << bits) - 1));
+}
+
+double R_unif_index(double dn)
+{
+    if (Sample_kind == ROUNDING)
+	return R_unif_index_0(dn);
+
+    // rejection sampling from integers below the next larger power of two
+    if (dn <= 0)
+	return 0.0;
+    int bits = (int) ceil(log2(dn));
+    double dv;
+    do { dv = rbits(bits); } while (dn <= dv);
+    return dv;
+}
+
+Sampletype R_sample_kind() { return Sample_kind; }
+
